import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

/**
 * <p>Represents a collection of {@link ServerResponse}s to be sent from the
 * server to one or more client(s) as the result of a user registration or
 * deregistration, or in response to a command received from some client. The
 * methods in {@link ServerModel} which correspond to these activities each
 * return an instance of this class. As outlined in the <a
 * href="../spec.shtml">protocol specification</a>, each one of these activities
 * results in at least one response being sent to a client.</p>
 *
 * <p>We will test each of the aforementioned {@code ServerModel} methods by
 * checking whether the {@code Broadcast}s they return exactly match the
 * reference {@code Broadcast} constructed in the test case, except for tests
 * that expect an error code, for which the error messages may be different.
 * See {@link ClientConnectionTest} for examples.</p>
 *
 * <p>You should not edit this file.</p>
 *
 * @see ServerModel
 * @see ServerResponse
 * @see ClientConnectionTest
 */
public final class Broadcast {
    private final Map<Integer, List<ServerResponse>> responses;

    /**
     * Constructs an empty {@code Broadcast}.
     */
    public Broadcast() {
        responses = new TreeMap<>();
    }

    /**
     * <p>Adds a {@code ServerResponse} to the list of responses to be
     * broadcast to the user with specified user ID. Note that the {@link
     * ServerResponse#getSender() prefix} of the command may be different from
     * the ID of the user receiving the command: it is often necessary to relay
     * commands sent by one client to another one. This method will not cause a
     * duplicate response to be sent to the client; if the given {@code
     * response} is already being sent to {@code userId} in this broadcast,
     * this method has no effect.</p>
     *
     * @param userId
     *         the ID of the user to send to
     * @param response
     *         the response to send
     */
    public void send(int userId, ServerResponse response) {
        if (!responses.containsKey(userId)) {
            responses.put(userId, new LinkedList<ServerResponse>());
        }
        List<ServerResponse> userResponses = responses.get(userId);
        if (!userResponses.contains(response)) {
            responses.get(userId).add(response);
        }
    }

    /**
     * A convenience method which constructs an {@code ERROR} response and
     * sends it to the user with specified user ID, exactly as if using {@link
     * #send(int, ServerResponse)}.
     *
     * @param userId
     *         the ID of the user to send to
     * @param senderNick
     *         the user's nick
     * @param errorCode
     *         the error code to send
     * @param errorMessage
     *         the (optional) error message to include
     */
    public void sendError(int userId, String senderNick,
            ServerErrorCode errorCode, String errorMessage) {
        String codeStr = Integer.toString(ServerErrorCode.getCode(errorCode));
        ServerResponse response = new ServerResponse(senderNick,
                Command.Type.ERROR, codeStr, null, errorMessage);
        send(userId, response);
    }

    /**
     * Appends all responses in {@code other} to this broadcast, excluding any
     * duplicates sent to the same user ID in both broadcasts. In short, merges
     * the two broadcasts together.
     *
     * @param other
     *         the broadcast to merge
     */
    public void sendAll(Broadcast other) {
        for (Map.Entry<Integer, List<ServerResponse>> entry : other.responses
                .entrySet()) {
            if (this.responses.containsKey(entry.getKey())) {
                List<ServerResponse> userResponses = this.responses
                        .get(entry.getKey());
                for (ServerResponse response : entry.getValue()) {
                    if (!userResponses.contains(response)) {
                        userResponses.add(response);
                    }
                }
            } else {
                this.responses.put(entry.getKey(), entry.getValue());
            }
        }
    }

    Map<Integer, List<ServerResponse>> getResponses() {
        return Collections.unmodifiableMap(responses);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        Broadcast broadcast = (Broadcast) o;

        if (!responses.equals(broadcast.responses)) {
            return false;
        }

        return true;
    }

    @Override
    public int hashCode() {
        return responses.hashCode();
    }

    @Override
    public String toString() {
        return responses.toString();
    }
}
