
import org.junit.Before;
import org.junit.Test;

import java.util.Collection;

import static org.junit.Assert.*;

public final class ClientConnectionTest {
    private ServerModel model;

    /*
     * If you annotate a method with [@Before], it will be called before
     * every test. You can use this to add set-up code, such as creating a
     * [ServerModel] as we do here.
     */
    @Before
    public void setUp() {
        model = new ServerModel();
    }

    /*
     * Test that no users are registered when the server model is
     * instantiated.
     */
    
    @Test
    public void testEmptyOnInit() {
        assertTrue("no registered users", model.getRegisteredUsers().isEmpty());
    }

    /*
     * Test that the [getRegisteredUsers] returns the proper output after one
     * user is registered.
     */
    
    @Test
    public void testRegisterSingleUser() {
        /*
         * In most of your tests, you will need to construct an expected
         * Broadcast to compare against the actual output from the model.
         * Here, we do that by instantiating a new [Broadcast] and then
         * sending the expected responses to each user ID. Be careful that
         * you use the proper user IDs and nicknames in your tests!
         */
        Broadcast expected = new Broadcast();
        expected.send(0,
                new ServerResponse("User0", Command.Type.CONNECT, null));


        /*
         * The definition of the [equals] method for [Broadcast] checks that
         * the two broadcasts send the same list of responses to each user ID.
         * Thus, to check that the model actually outputs the correct Broadcast,
         * we can just [assertEquals] against the expected value we made above.
         */
        assertEquals("broadcast", expected, model.registerUser(0));

        /*
         * In addition to checking that [registerUser] returns the correct
         * [Broadcast], we use the testing methods of the [ServerModel] to
         * verify some internal properties of the model.
         */
        Collection<String> registeredUsers = model.getRegisteredUsers();
        assertEquals("num. registered users", 1, registeredUsers.size());
        assertTrue("User0 is registered", registeredUsers.contains("User0"));
    }

    /*
     * Test that the [getRegisteredUsers] returns the proper output after
     * multiple users are registered.
     */
    
    @Test
    public void testRegisterMultipleUsers() {
        /*
         * Here, we see a slightly more complicated test. Again, note the
         * pattern where we construct an expected output and then compare it
         * against the output of [registerUser].
         *
         * This test working is INCREDIBLY IMPORTANT, as our tests on the server
         * rely on the specification of how usernames are assigned upon
         * registration.
         */
        Broadcast expected0 = new Broadcast();
        expected0.send(0,
                new ServerResponse("User0", Command.Type.CONNECT, null));
        assertEquals("broadcast", expected0, model.registerUser(0));

        Broadcast expected1 = new Broadcast();
        expected1.send(1,
                new ServerResponse("User1", Command.Type.CONNECT, null));
        assertEquals("broadcast", expected1, model.registerUser(1));

        Broadcast expected2 = new Broadcast();
        expected2.send(2,
                new ServerResponse("User2", Command.Type.CONNECT, null));
        assertEquals("broadcast", expected2, model.registerUser(2));

        Collection<String> registeredUsers = model.getRegisteredUsers();
        assertEquals("num. registered users", 3, registeredUsers.size());
        assertTrue("User0 is registered", registeredUsers.contains("User0"));
        assertTrue("User1 is registered", registeredUsers.contains("User1"));
        assertTrue("User2 is registered", registeredUsers.contains("User2"));
    }

    /*
     * Test that [deregisterUser] accurately updates the output of
     * [getRegisteredUsers].
     */
    
    @Test
    public void testDeregisterSingleUserErasesFromRegisteredUsers() {
        model.registerUser(0);
        model.deregisterUser(0);
        assertTrue("no users registered", model.getRegisteredUsers().isEmpty());
    }

    /*
     * Test that the server handles a NICK sent by someone not in any
     * channels correctly.
     */
    
    @Test
    public void testNickNotInChannels() {
        model.registerUser(0);

        /*
         * Here, we use the convenience method [makeNick] from [TestUtil] to
         * construct a NICK command as one might get from a client. You
         * should look at the similar methods we provide for each of the
         * command types.
         */
        Command nickCommand = ServerUtil.makeNick("cis120");
        /*
         * Here is an example of how to generate a [ServerResponse] from a
         * given [Command].
         */
        ServerResponse nickResponse = new ServerResponse("User0", nickCommand);

        Broadcast expected = new Broadcast();
        expected.send(0, nickResponse);
        assertEquals("broadcast", expected, model.handle(0, nickCommand));

        assertFalse("old nick no longer registered",
                model.getRegisteredUsers().contains("User0"));
        assertTrue("new nick now registered",
                model.getRegisteredUsers().contains("cis120"));
    }

    /*
     * Test that the server throws an error when a client attempts to change
     * his/her nick to one that is already in use.
     */
    
    @Test
    public void testNickCollision() {
        model.registerUser(0);
        model.registerUser(1);

        Command nickCommand = ServerUtil.makeNick("User1");
        TestUtil.assertErrorToUser(0, ServerErrorCode.NICK_ALREADY_IN_USE,
                model.handle(0, nickCommand));

        assertTrue("old nick still registered",
                model.getRegisteredUsers().contains("User0"));
        assertTrue("other user still registered",
                model.getRegisteredUsers().contains("User1"));
    }

    /*
     * Test that the server throws an error when the client attempts to
     * change his/her nick to be the same nick.
     */
    
    @Test
    public void testNickRedundant() {
        model.registerUser(0);

        Command nickCommand = ServerUtil.makeNick("User0");
        TestUtil.assertErrorToUser(0, ServerErrorCode.NICK_ALREADY_IN_USE,
                model.handle(0, nickCommand));

        assertEquals("num. registered users", 1,
                model.getRegisteredUsers().size());
        assertTrue("old nick still registered",
                model.getRegisteredUsers().contains("User0"));
    }

    /*
     * Test that the server properly generates new nicks for connected users;
     * specifically, if User0 changes his/her nick to something else, the
     * next person to connect to the server should be assigned User0 as a nick.
     */
    
    @Test
    public void testNickCollisionOnConnect() {
        model.registerUser(0);
        model.handle(0, ServerUtil.makeNick("User1"));

        ServerResponse connectResponse = new ServerResponse("User0",
                ServerUtil.makeConnect());

        Broadcast expected = new Broadcast();
        expected.send(1, connectResponse);
        assertEquals("broadcast", expected, model.registerUser(1));

        assertEquals("num. registered users", 2,
                model.getRegisteredUsers().size());
        assertTrue("User0 registered",
                model.getRegisteredUsers().contains("User0"));
        assertTrue("User1 registered",
                model.getRegisteredUsers().contains("User1"));
    }

    /*
     * Test that the server throws an error on nicks containing
     * non-alphanumeric characters.
     */
    
    @Test
    public void testNickInvalidParameter() {
        model.registerUser(0);

        Command nickCommand = ServerUtil.makeNick("@Imposter");
        TestUtil.assertErrorToUser(0, ServerErrorCode.INVALID_PARAMETER,
                model.handle(0, nickCommand));

        assertEquals("num. registered users", 1,
                model.getRegisteredUsers().size());
        assertTrue("old nick still registered",
                model.getRegisteredUsers().contains("User0"));
    }
}

