/**
 * <p>Represents a command string sent from a client to the server, after being
 * parsed into a more convenient data structure. It includes fields for the
 * three parts of these commands, the type, the two parameters (which may or
 * may not be present), and the payload (which also may not be present). See
 * the <a href="../spec.shtml">protocol specification</a> for more details
 * about the command format.</p>
 *
 * <p>You should not edit this file.</p>
 */
public final class Command {
    /**
     * This enumeration defines the various types of commands that the clients
     * and server operate with. Note that several of these can only be
     * generated by the server and will never be sent by a conforming client.
     */
    public static enum Type {
        CONNECT,
        NICK,
        CREATE,
        JOIN,
        NAMES,
        MESG,
        LEAVE,
        KICK,
        QUIT,
        INVITE,
        ERROR;

        public static int getNumParams(Type type) {
            switch (type) {
                case CONNECT:
                case QUIT:
                    return 0;
                case CREATE:
                case KICK:
                case INVITE:
                    return 2;
                default:
                    return 1;
            }
        }

        public static boolean allowsPayload(Type type) {
            switch (type) {
                case MESG:
                case ERROR:
                case LEAVE:
                case KICK:
                case NAMES:
                    return true;
                default:
                    return false;
            }
        }

        public static boolean requiresPayload(Type type) {
            switch (type) {
                case NAMES:
                case MESG:
                    return true;
                default:
                    return false;
            }
        }
    }

    /**
     * The command type. May never be {@code null}.
     */
    private final Type type;

    /**
     * The first parameter. May be {@code null} if absent.
     */
    private final String param0;

    /**
     * The second parameter. May be {@code null} if absent.
     */
    private final String param1;

    /**
     * The payload. May be {@code null} if absent.
     */
    private final String payload;

    /**
     * Constructs a {@code Command} with the provided components. The parameter
     * count and the presence or absence of a payload must agree with the
     * command type according to the <a href="../spec.shtml">specification</a>.
     *
     * @param type
     *         the command type
     * @param param0
     *         the first parameter, or {@code null} if absent
     * @param param1
     *         the second parameter, or {@code null} if absent
     * @param payload
     *         the payload, or {@code null} if absent
     * @throws IllegalArgumentException
     *         if {@code type} is {@code null}, or if the command is
     *         syntactically invalid; it has an incorrect parameter count, it
     *         is missing a payload where one is required, or it has a payload
     *         where one is invalid
     */
    public Command(Type type, String param0, String param1, String payload) {
        if (type == null) {
            throw new IllegalArgumentException("type must be non-null");
        }

        this.type = type;
        this.param0 = param0;
        this.param1 = param1;
        this.payload = payload;

        if (this.param0 == null && this.param1 != null) {
            throw new IllegalArgumentException(
                    "param1 should be null when param0 is null, in command " +
                            this);
        }
        if (getParameterCount() != Type.getNumParams(this.type)) {
            throw new IllegalArgumentException(
                    "Invalid number of parameters for command " + this);
        }
        if (payload != null && !Type.allowsPayload(type) ||
                payload == null && Type.requiresPayload(type)) {
            throw new IllegalArgumentException(
                    "Invalid payload for command " + this);
        }
    }

    /**
     * Convenience constructor for {@code Command}s with no parameters.
     *
     * @param type
     *         the command type
     * @param payload
     *         the payload, or {@code null} if absent
     * @see #Command(Type, String, String, String)
     */
    public Command(Type type, String payload) {
        this(type, null, null, payload);
    }

    /**
     * Returns the command type.
     *
     * @return the command type
     */
    public Type getType() {
        return type;
    }

    /**
     * Returns the specified parameter.
     *
     * @param index
     *         the 0-based index of the parameter to retrieve
     * @return the parameter at the specified index
     * @throws IllegalArgumentException
     *         if {@code index} is invalid
     */
    public String getParameter(int index) {
        if (index < 0 || index >= getParameterCount()) {
            throw new IllegalArgumentException(
                    "Cannot access parameter " + index + " from command " +
                            this);
        }
        if (index == 0) {
            return this.param0;
        } else {
            return this.param1;
        }
    }

    /**
     * Counts the number of parameters in this command.
     *
     * @return the parameter count
     */
    public int getParameterCount() {
        if (param0 == null) {
            return 0;
        } else if (param1 == null) {
            return 1;
        } else {
            return 2;
        }
    }

    /**
     * Returns the payload.
     *
     * @return the payload, or {@code null} if there is none
     */
    public String getPayload() {
        return payload;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        Command command = (Command) o;

        if (type != command.type) {
            return false;
        }
        if (param0 != null ? !param0.equals(command.param0)
                : command.param0 != null) {
            return false;
        }
        if (param1 != null ? !param1.equals(command.param1)
                : command.param1 != null) {
            return false;
        }
        if (payload != null ? !payload.equals(command.payload)
                : command.payload != null) {
            return false;
        }

        return true;
    }

    @Override
    public int hashCode() {
        int result = type.hashCode();
        result = 31 * result + (param0 != null ? param0.hashCode() : 0);
        result = 31 * result + (param1 != null ? param1.hashCode() : 0);
        result = 31 * result + (payload != null ? payload.hashCode() : 0);
        return result;
    }

    /**
     * Joins this {@code Command}'s parameters in a space-delimited string.
     *
     * @return the space-delimited parameters
     */
    private String joinParameters() {
        StringBuilder builder = new StringBuilder();
        if (param0 != null) {
            builder.append(param0);
        }
        if (param1 != null) {
            builder.append(" ");
            builder.append(param1);
        }
        return builder.toString();
    }

    /**
     * Returns the representation of the {@code Command} as it would be sent
     * over the wire to the server.
     *
     * @return the network representation of the command
     */
    @Override
    public String toString() {
        String argStr = joinParameters();
        boolean noParams = getParameterCount() == 0;

        if (noParams && payload == null) {
            return type.toString();
        } else if (noParams) {
            return String.format("%s :%s", type, payload);
        } else if (payload == null) {
            return String.format("%s %s", type, argStr);
        } else {
            return String.format("%s %s :%s", type, argStr, payload);
        }
    }
}
