import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingQueue;

/**
 * <p>Backend code for the server which sets up and handles network connections
 * from clients. The backend queries the {@link ServerModel} whenever a client
 * connects, disconnects, or sends a message, and it sends the resulting
 * {@link Broadcast} to the appropriate parties over the network.</p>
 *
 * <p>The backend can be run using {@link ServerMain}, not this file. You do
 * not need to read or modify this file for the purposes of this
 * assignment.</p>
 */

final class ServerBackend implements Runnable {
    /*
     * The ServerModel is NOT thread-safe; it should only be touched on the
     * model thread after being initialized.
     */
    private final ServerModel model;

    private final BlockingQueue<Action> actionQueue;

    private volatile ServerSocket serverSocket;

    private final Map<Integer, Socket> openSockets;

    private volatile boolean running;

    private volatile Thread modelThread;

    public ServerBackend(ServerModel model) {
        if (model == null) {
            throw new NullPointerException();
        }
        this.model = model;
        actionQueue = new LinkedBlockingQueue<>();
        serverSocket = null;
        openSockets = Collections
                .synchronizedMap(new HashMap<Integer, Socket>());
        running = false;
        modelThread = null;
    }

    public boolean isRunning() {
        return running;
    }

    @Override
    public void run() {
        running = true;

        // attempt to open the ServerSocket, abort on failure
        try {
            serverSocket = new ServerSocket(21212);
        } catch (IOException iox) {
            iox.printStackTrace();
            running = false;
            serverSocket = null;
        }

        // start the model thread
        modelThread = new Thread(new Runnable() {
            @Override
            public void run() {
                while (running || !actionQueue.isEmpty()) {
                    Action action;
                    try {
                        action = actionQueue.take();
                    } catch (InterruptedException ix) {
                        continue;
                    }
                    // be resilient against any errors that occur in the model
                    // implementation
                    try {
                        doBroadcast(action.run());
                    } catch (RuntimeException rx) {
                        rx.printStackTrace();
                    }
                }
                try {
                    if (serverSocket != null && !serverSocket.isClosed()) {
                        serverSocket.close();
                    }
                } catch (IOException iox) {
                    iox.printStackTrace();
                }
            }
        }, "Model thread");
        modelThread.start();

        // await new connections on the current thread
        ExecutorService workerPool = Executors.newCachedThreadPool();
        try {
            int nextId = 0;
            while (running && !serverSocket.isClosed()) {
                int userId = nextId++;
                Socket clientSocket = serverSocket.accept();
                openSockets.put(userId, clientSocket);
                actionQueue.add(new Registration(userId));
                workerPool.execute(new ConnectionWorker(userId, clientSocket));
            }
        } catch (IOException iox) {
            iox.printStackTrace();
        } finally {
            running = false;
            workerPool.shutdown();
            try {
                if (serverSocket != null && !serverSocket.isClosed()) {
                    serverSocket.close();
                }
            } catch (IOException iox) {
                iox.printStackTrace();
            } finally {
                serverSocket = null;
            }
            synchronized (openSockets) {
                Iterator<Socket> iterator = openSockets.values().iterator();
                while (iterator.hasNext()) {
                    Socket clientSocket = iterator.next();
                    try {
                        clientSocket.close();
                    } catch (IOException iox) {
                        iox.printStackTrace();
                    } finally {
                        iterator.remove();
                    }
                }
            }
        }
    }

    private void doBroadcast(Broadcast broadcast) {
        for (int userId : broadcast.getResponses().keySet()) {
            try {
                Socket clientSocket = openSockets.get(userId);
                PrintWriter pw = new PrintWriter(
                        clientSocket.getOutputStream());
                for (ServerResponse response : broadcast.getResponses()
                        .get(userId)) {
                    pw.println(response.toString());
                    System.out.printf("Response sent to user %d: \"%s\"\n", userId,
                            response.toString());
                }
                pw.flush();
            } catch (IOException iox) {
                iox.printStackTrace();
            }
        }
    }

    public void stop() {
        running = false;
        if (serverSocket != null && !serverSocket.isClosed()) {
            try {
                serverSocket.close();
            } catch (IOException iox) {
                iox.printStackTrace();
            }
        }
        if (modelThread != null) {
            modelThread.interrupt();
        }
    }

    private final class ConnectionWorker implements Runnable {
        private final int userId;

        private final Socket clientSocket;

        public ConnectionWorker(int userId, Socket clientSocket) {
            this.userId = userId;
            this.clientSocket = clientSocket;
        }

        @Override
        public void run() {
            try (BufferedReader reader = new BufferedReader(
                    new InputStreamReader(clientSocket.getInputStream()))) {
                while (running && !clientSocket.isClosed()) {
                    String line = reader.readLine();
                    if (line != null) {
                        System.out.printf("Command string received from user %d: \"%s\"\n",
                                userId, line);
                        String payload;
                        if (line.startsWith(":")) {
                            int index = line.indexOf(' ');
                            payload = line.substring(index + 1);
                        } else {
                            payload = line;
                        }
                        Message message = new Message(userId, payload);
                        actionQueue.add(message);
                    } else {
                        clientSocket.close();
                        actionQueue.add(new Disconnection(userId));
                    }
                }
            } catch (IOException iox) {
                iox.printStackTrace();
                actionQueue.add(new Disconnection(userId));
            } finally {
                openSockets.remove(userId);
            }
        }
    }

    private static interface Action {
        Broadcast run();
    }

    private final class Registration implements Action {
        private final int userId;

        public Registration(int userId) {
            this.userId = userId;
        }

        @Override
        public Broadcast run() {
            return model.registerUser(userId);
        }
    }

    private final class Disconnection implements Action {
        private final int userId;

        public Disconnection(int userId) {
            this.userId = userId;
        }

        @Override
        public Broadcast run() {
            return model.deregisterUser(userId);
        }
    }

    private final class Message implements Action {
        private final int userId;

        private final String payload;

        public Message(int userId, String payload) {
            this.userId = userId;
            this.payload = payload;
        }

        @Override
        public Broadcast run() {
            Command command = CommandParser.parse(payload);
            return model.handle(userId, command);
        }
    }
}
