/**
 * <p>An enum to represent the various categories of error which are specified
 * by the protocol. Each enum constant corresponds to a particular error code
 * which is transmitted to the client as the first parameter of an {@code
 * ERROR} command.</p>
 *
 * <p>You should not edit this file.</p>
 */
public enum ServerErrorCode {
    /**
     * Sent by the server in cases when it receives a command that it should
     * not received per the protocol specification.
     */
    COMMAND_UNSUPPORTED_BY_SERVER,

    /**
     * Response by the server when it encounters a parameter value that is
     * illegal for the given command.
     */
    INVALID_PARAMETER,

    /**
     * Response by the server when the channel specified in a command is not
     * recognized by the server.
     */
    CHANNEL_DOES_NOT_EXIST,

    /**
     * Response by the server when the target specified in the command is not
     * registered on the server.
     */
    TARGET_DOES_NOT_EXIST,

    /**
     * Response by the server when a client attempts to send a command
     * requiring the sender to be in the specified channel to a channel in
     * which he or she is not a member.
     */
    SENDER_NOT_IN_CHANNEL,

    /**
     * Response by the server when a client attempts to send a command that
     * requires the target to be in particular channel, but the target is not.
     */
    TARGET_NOT_IN_CHANNEL,

    /**
     * Response by the server when a client which does not own a channel
     * attempts to send a command that is restricted to the owner; currently,
     * the only such command is KICK.
     */
    SENDER_NOT_OWNER,

    /**
     * Response by the server when a client attempts to join a channel to which
     * he or she has not been invited.
     */
    SENDER_NOT_INVITED,

    /**
     * Response by the server when a client attempts to issue an invite to a
     * channel that is not invite-only.
     */
    INVITE_TO_PUBLIC_CHANNEL,

    /**
     * Response by the server when a client attempts to change his or her nick
     * to a nickname that is already in use by another user.
     */
    NICK_ALREADY_IN_USE,

    /**
     * Response by the server when a client attempts to create a channel whose
     * name is already used by another channel on the server.
     */
    CHANNEL_ALREADY_EXISTS;

    /**
     * Returns the numeric error code corresponding to the specified {@code
     * ServerErrorCode} enum constant.
     *
     * @param errorCode
     *         the error code to look up
     * @return the numeric error code
     */
    public static int getCode(ServerErrorCode errorCode) {
        switch (errorCode) {
            case COMMAND_UNSUPPORTED_BY_SERVER:
                return 400;
            case INVALID_PARAMETER:
                return 401;
            case CHANNEL_DOES_NOT_EXIST:
                return 402;
            case TARGET_DOES_NOT_EXIST:
                return 403;
            case SENDER_NOT_IN_CHANNEL:
                return 404;
            case TARGET_NOT_IN_CHANNEL:
                return 405;
            case SENDER_NOT_OWNER:
                return 406;
            case SENDER_NOT_INVITED:
                return 407;
            case INVITE_TO_PUBLIC_CHANNEL:
                return 408;
            case NICK_ALREADY_IN_USE:
                return 500;
            case CHANNEL_ALREADY_EXISTS:
                return 501;
            default:
                throw new IllegalArgumentException("unknown error code");
        }
    }
}
