import org.junit.Before;
import org.junit.Test;

import static org.junit.Assert.*;

/**
 * <p>Use this class to write your own unit tests for {@link ServerModel}. You
 * can use the tests in {@link ClientConnectionTest} as a guide.</p>
 *
 * @see ServerModel
 * @see ClientConnectionTest
 */
public final class ServerModelTest {
    private ServerModel model;

    /*
     * As in [ClientConnectionTest], we initialize a fresh [ServerModel] before
     * every test so that tests don't affect each other.
     */
    @Before
    public void setUp() {
        model = new ServerModel();
    }

    @Test
    public void testNickBroadcastsToChannelWhereMember() {
        model.registerUser(0);
        model.registerUser(1);

        model.handle(0, ServerUtil.makeCreate("java", false));
        model.handle(1, ServerUtil.makeJoin("java"));

        Command nickCommand = ServerUtil.makeNick("Duke");

        Broadcast expected = new Broadcast();
        expected.send(0, new ServerResponse("User0", nickCommand));
        expected.send(1, new ServerResponse("User0", nickCommand));
        assertEquals("broadcast", expected, model.handle(0, nickCommand));

        assertFalse("old nick not in channel",
                model.getUsers("java").contains("User0"));
        assertTrue("new nick is in channel",
                model.getUsers("java").contains("Duke"));
        assertTrue("unaffected user still in channel",
                model.getUsers("java").contains("User1"));
    }

    @Test
    public void testDeregisterSendsQuitWhereMember() {
        model.registerUser(0);
        model.registerUser(1);
        model.handle(0, ServerUtil.makeCreate("java", false));
        model.handle(1, ServerUtil.makeJoin("java"));

        Command quitCommand = ServerUtil.makeQuit();
        Broadcast expected = new Broadcast();
        expected.send(0, new ServerResponse("User1", quitCommand));
        assertEquals("broadcast", expected, model.deregisterUser(1));

        assertTrue("channel still exists",
                model.getChannels().contains("java"));
        assertEquals("num. users in channel", 1,
                model.getUsers("java").size());
        assertTrue("unaffected user still in channel",
                model.getUsers("java").contains("User0"));
    }

    @Test
    public void testCreateNewChannel() {
        model.registerUser(0);

        Command createCommand = ServerUtil.makeCreate("java", false);
        Broadcast expected = new Broadcast();
        expected.send(0, new ServerResponse("User0", createCommand));

        assertEquals("broadcast", expected, model.handle(0, createCommand));

        assertTrue("channel exists", model.getChannels().contains("java"));
        assertTrue("channel has creator",
                model.getUsers("java").contains("User0"));
        assertEquals("channel has owner", "User0", model.getOwner("java"));
    }

    @Test
    public void testJoinChannelExistsNotMember() {
        model.registerUser(0);
        model.registerUser(1);

        Command joinCommand = ServerUtil.makeJoin("java");
        Command namesCommand = ServerUtil.makeNames("java", "@User0 User1");

        Broadcast expected = new Broadcast();
        expected.send(0, new ServerResponse("User1", joinCommand));
        expected.send(1, new ServerResponse("User1", joinCommand));
        expected.send(1, new ServerResponse("User1", namesCommand));

        model.handle(0, ServerUtil.makeCreate("java", false));
        assertEquals("broadcast", expected, model.handle(1, joinCommand));

        assertTrue("User0 in channel",
                model.getUsers("java").contains("User0"));
        assertTrue("User1 in channel",
                model.getUsers("java").contains("User1"));
        assertEquals("num. users in channel", 2,
                model.getUsers("java").size());
    }

    @Test
    public void testLeaveChannelExistsMember() {
        model.registerUser(0);
        model.registerUser(1);

        Command joinCommand = ServerUtil.makeJoin("java");
        Command leaveCommand = ServerUtil.makeLeave("java", "bye fools");

        Broadcast expected = new Broadcast();
        expected.send(0, new ServerResponse("User1", leaveCommand));
        expected.send(1, new ServerResponse("User1", leaveCommand));

        model.handle(0, ServerUtil.makeCreate("java", false));
        model.handle(1, joinCommand);
        assertEquals("broadcast", expected, model.handle(1, leaveCommand));

        assertTrue("User0 in channel",
                model.getUsers("java").contains("User0"));
        assertFalse("User1 not in channel",
                model.getUsers("java").contains("User1"));
        assertEquals("num. users in channel", 1,
                model.getUsers("java").size());
    }

    @Test
    public void testMesgChannelExistsMember() {
        model.registerUser(0);
        model.registerUser(1);

        Command mesgCommand = ServerUtil.makeMesg("dinosaur", "hey kids");

        Broadcast expected = new Broadcast();
        expected.send(0, new ServerResponse("User1", mesgCommand));
        expected.send(1, new ServerResponse("User1", mesgCommand));

        model.handle(0, ServerUtil.makeCreate("dinosaur", false));
        model.handle(1, ServerUtil.makeJoin("dinosaur"));
        assertEquals("broadcast", expected, model.handle(1, mesgCommand));
    }

    @Test
    public void testKickOneChannel() {
        model.registerUser(0);
        model.registerUser(1);
        model.registerUser(2);

        model.handle(0, ServerUtil.makeCreate("java", false));
        model.handle(1, ServerUtil.makeJoin("java"));
        model.handle(2, ServerUtil.makeJoin("java"));

        Command kickCommand = ServerUtil.makeKick("java", "User2", "womp womp");
        Broadcast expected = new Broadcast();
        expected.send(0, new ServerResponse("User0", kickCommand));
        expected.send(1, new ServerResponse("User0", kickCommand));
        expected.send(2, new ServerResponse("User0", kickCommand));
        assertEquals(expected, model.handle(0, kickCommand));

        assertEquals("num. users in channel", 2,
                model.getUsers("java").size());
        assertTrue("User0 still in channel",
                model.getUsers("java").contains("User0"));
        assertTrue("User1 still in channel",
                model.getUsers("java").contains("User1"));
    }

    @Test
    public void testNamesUnsupportedOperation() {
        model.registerUser(0);

        Command namesCommand = ServerUtil
                .makeNames("java", "@User0 User1 User2 User4");

        TestUtil.assertErrorToUser(0,
                ServerErrorCode.COMMAND_UNSUPPORTED_BY_SERVER,
                model.handle(0, namesCommand));
    }


    @Test
    public void testInviteByOwnerFollowedByJoin() {
        model.registerUser(0);
        model.registerUser(1);

        model.handle(0, ServerUtil.makeCreate("java", true));
        model.handle(0, ServerUtil.makeInvite("java", "User1"));

        Command joinCommand = ServerUtil.makeJoin("java");
        ServerResponse joinResponse = new ServerResponse("User1", joinCommand);
        ServerResponse namesResponse = new ServerResponse("User1",
                ServerUtil.makeNames("java", "@User0 " + "User1"));

        Broadcast expected = new Broadcast();
        expected.send(0, joinResponse);
        expected.send(1, joinResponse);
        expected.send(1, namesResponse);

        assertEquals("broadcast", expected, model.handle(1, joinCommand));
        assertEquals("num. users in channel", 2,
                model.getUsers("java").size());
        assertTrue("User0 in channel",
                model.getUsers("java").contains("User0"));
        assertTrue("User1 in channel",
                model.getUsers("java").contains("User1"));
    }

    @Test
    public void testInviteOnlyPossibleByOwner() {
        model.registerUser(0);
        model.registerUser(1);
        model.registerUser(2);

        model.handle(0, ServerUtil.makeCreate("java", true));
        model.handle(0, ServerUtil.makeInvite("java", "User1"));
        model.handle(1, ServerUtil.makeJoin("java"));

        TestUtil.assertErrorToUser(1, ServerErrorCode.SENDER_NOT_OWNER,
                model.handle(1, ServerUtil.makeInvite("java", "User2")));

        // Verify that invite was not issued.
        Command joinCommand = ServerUtil.makeJoin("java");
        TestUtil.assertErrorToUser(2, ServerErrorCode.SENDER_NOT_INVITED,
                model.handle(2, joinCommand));
    }
}
