/**
 * <p>Represents responses sent from the server to the client in response to a
 * client {@link Command}, a user registration, or a user deregistration. A
 * {@code ServerResponse} consists of a base {@code Command} in the same form
 * as those sent from the client to server, along with an additional sender
 * which indicates the <em>sender</em> whose action generated the response.</p>
 *
 * <p>There are generally two use cases for this class: first, for responses
 * that involve relaying a client {@code Command} back to the sender, and
 * possibly also to other clients; and second, for responses like {@code ERROR}
 * that are generated by the server. See the <a href="../spec.shtml">protocol
 * specification</a> for more details.</p>
 *
 * <p>You should not edit this file.</p>
 */
public final class ServerResponse {
    private final String sender;

    private final Command command;

    /**
     * Constructs a {@code ServerResponse} from a base {@link Command} and the
     * {@code sender}.
     *
     * @param sender
     *         the sender whose action generated this response
     * @param command
     *         the base command
     * @throws IllegalArgumentException
     *         if {@code sender} or {@code command} is {@code null}
     */
    public ServerResponse(String sender, Command command) {
        if (sender == null || command == null) {
            throw new IllegalArgumentException(
                    "ServerResponse must have non-null components");
        }
        this.sender = sender;
        this.command = command;
    }

    /**
     * A convenience constructor which builds a new {@link Command} from its
     * components to create the {@code ServerResponse}.
     *
     * @param sender
     *         the sender whose action generated this response
     * @param commandType
     *         the command type
     * @param p0
     *         the first parameter, or {@code null} if absent
     * @param p1
     *         the second parameter, or {@code null} if absent
     * @param payload
     *         the payload, or {@code null} if absent
     * @throws IllegalArgumentException
     *         if {@code sender} is {@code null}, or if the command is
     *         syntactically invalid; see {@code Command} for details
     * @see Command#Command(Command.Type, String, String, String)
     */
    public ServerResponse(String sender, Command.Type commandType, String p0,
            String p1, String payload) {
        this(sender, new Command(commandType, p0, p1, payload));
    }

    /**
     * A convenience constructor which builds a new {@link Command} without
     * parameters from its (other) components to create the {@code
     * ServerResponse}.
     *
     * @param sender
     *         the sender whose action generated this response
     * @param commandType
     *         the command type
     * @param payload
     *         the payload, or {@code null} if absent
     * @throws IllegalArgumentException
     *         if {@code sender} is {@code null}, or if the command is
     *         syntactically invalid; see {@code Command} for details
     * @see Command#Command(Command.Type, String)
     */
    public ServerResponse(String sender, Command.Type commandType,
            String payload) {
        this(sender, new Command(commandType, payload));
    }

    /**
     * @return the sender
     */
    public String getSender() {
        return sender;
    }

    /**
     * @return the base command
     */
    public Command getCommand() {
        return command;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        ServerResponse that = (ServerResponse) o;

        if (!command.equals(that.command)) {
            return false;
        }
        if (!sender.equals(that.sender)) {
            return false;
        }

        return true;
    }

    @Override
    public int hashCode() {
        int result = sender.hashCode();
        result = 31 * result + command.hashCode();
        return result;
    }

    @Override
    public String toString() {
        return ":" + sender + " " + command.toString();
    }
}
