import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

/**
 * <p>Class providing static utility methods to help in creating and testing
 * the {@link ServerModel}.</p>
 */
public final class ServerUtil {
    /**
     * Convenience method to generate a {@code CONNECT} {@code Command}.
     *
     * @return a new {@code CONNECT} {@code Command}
     */
    public static Command makeConnect() {
        return new Command(Command.Type.CONNECT, null);
    }

    /**
     * Convenience method to generate a {@code NICK} {@code Command}.
     *
     * @param newNick
     *         the sender's new desired nickname
     * @return a new {@code NICK} {@code Command} with the specified component
     */
    public static Command makeNick(String newNick) {
        return new Command(Command.Type.NICK, newNick, null, null);
    }

    /**
     * Convenience method to generate a {@code CREATE} {@code Command}.
     *
     * @param channelName
     *         the name of the channel to be created
     * @param inviteOnly
     *         {@code true} if the channel should be invite-only, and {@code
     *         false} otherwise
     * @return a new {@code CREATE} {@code Command} with the specified
     * components
     */
    public static Command makeCreate(String channelName, boolean inviteOnly) {
        return new Command(Command.Type.CREATE, channelName,
                inviteOnly ? "1" : "0", null);
    }

    /**
     * Convenience method to generate a {@code JOIN} {@code Command}.
     *
     * @param channelName
     *         the name of the channel to be joined
     * @return a new {@code JOIN} {@code Command} with the specified component
     */
    public static Command makeJoin(String channelName) {
        return new Command(Command.Type.JOIN, channelName, null, null);
    }

    /**
     * Convenience method to generate a {@code NAMES} {@code Command}.
     *
     * @param channelName
     *         the name of the channel whose member names were requested
     * @param payload
     *         the payload representing the names of the channel users. This
     *         should follow the payload format specified in the assignment
     *         instructions
     * @return a new {@code NAMES} {@code Command} with the specified components
     */
    public static Command makeNames(String channelName, String payload) {
        return new Command(Command.Type.NAMES, channelName, null, payload);
    }

    /**
     * Convenience method to generate a {@code LEAVE} {@code Command}.
     *
     * @param channelName
     *         the name of the channel to leave
     * @param message
     *         the (optional) leave message
     * @return a new {@code LEAVE} {@code Command} with the specified components
     */
    public static Command makeLeave(String channelName, String message) {
        return new Command(Command.Type.LEAVE, channelName, null, message);
    }

    /**
     * Convenience method to generate a {@code MESG} {@code Command}.
     *
     * @param channelName
     *         the name of the channel on which to send the message
     * @param payload
     *         the payload representing the message sent to the channel
     * @return a new {@code MESG} {@code Command} with the specified components
     */
    public static Command makeMesg(String channelName, String payload) {
        return new Command(Command.Type.MESG, channelName, null, payload);
    }

    /**
     * Convenience method to generate a {@code KICK} {@code Command}.
     *
     * @param channelName
     *         the name of the channel from which to kick the target
     * @param target
     *         the nick of the user to kick
     * @param payload
     *         the (optional) kick message
     * @return a new {@code KICK} {@code Command} with the specified components
     */
    public static Command makeKick(String channelName, String target,
            String payload) {
        return new Command(Command.Type.KICK, channelName, target, payload);
    }

    /**
     * Convenience method to generate an {@code INVITE} {@code Command}.
     *
     * @param channelName
     *         the name of the channel to which the target is invited
     * @param invitee
     *         the nick of the user to invite
     * @return a new {@code INVITE} {@code Command} with the specified
     * components
     */
    public static Command makeInvite(String channelName, String invitee) {
        return new Command(Command.Type.INVITE, channelName, invitee, null);
    }

    /**
     * Convenience method to generate a {@code QUIT} {@code Command}.
     *
     * @return a new {@code QUIT} {@code Command}
     */
    public static Command makeQuit() {
        return new Command(Command.Type.QUIT, null);
    }

    /**
     * Generates a correct payload for a {@code NAMES} response for some
     * channel given the owner nick for that channel and a collection of all
     * users on that channel, including the owner.
     *
     * @param owner
     *         the channel owner's nick
     * @param nicks
     *         a collection of nicks for users of the channel, including the
     *         owner
     * @return a correct {@code NAMES} payload containing the nicks specified
     * @throws IllegalArgumentException
     *         if {@code owner} or {@code nicks} is {@code null}, or of {@code
     *         nicks} does not contain {@code owner}
     */
    public static String generateNamesPayload(String owner,
            Collection<String> nicks) {
        if (owner == null || nicks == null || !nicks.contains(owner)) {
            throw new IllegalArgumentException(
                    "bad argument to generateNamesPayload");
        }

        List<String> nicksList = new LinkedList<>(nicks);
        Collections.sort(nicksList);
        String payload = "";
        for (String nick : nicks) {
            if (nick.equals(owner)) {
                payload += "@" + nick;
            } else {
                payload += nick;
            }
            payload += " ";
        }
        return payload.trim();
    }

    /**
     * Determines if a given nickname is valid (contains 1 or more alphanumeric
     * characters).
     *
     * @param nick
     *         the string of the nickname to determine if it is valid
     * @return {@code true} if the string is a valid nickname
     */
    public static boolean isValidNickname(String nick) {
        if (nick == null || nick.isEmpty()) {
            return false;
        }
        for (char c : nick.toCharArray()) {
            if (!Character.isLetterOrDigit(c)) {
                return false;
            }
        }
        return true;
    }

    /**
     * Prevents the instantiation of any {@code ServerUtil} objects, which
     * would be nonsensical.
     */
    private ServerUtil() {
    }
}
