import java.util.Collections;
import java.util.List;
import java.util.Set;

import static org.junit.Assert.*;

/**
 * <p>Class providing static utility methods to help in testing the 
 * {@link ServerModel}</p>
 */
public final class TestUtil {
    /**
     * When several error codes apply to an invalid command, your
     * {@code ServerModel} may send any one of them in reply. This helper
     * method checks whether one of some set of error codes was sent to the
     * user in the given broadcast; in addition, it verifies that no additional
     * responses were sent to any user in the broadcast.
     *
     * @param userId
     *         the user's ID
     * @param expected
     *         the set of allowed error codes
     * @param actual
     *         the broadcast to check against
     */
    public static void assertErrorToUser(int userId,
            Set<ServerErrorCode> expected, Broadcast actual) {
        assertEquals("num. users in broadcast", 1,
                actual.getResponses().size());
        assertTrue("user ID in broadcast",
                actual.getResponses().containsKey(userId));
        List<ServerResponse> sent = actual.getResponses().get(userId);
        assertEquals("num. responses in broadcast", 1, sent.size());
        ServerResponse response = sent.get(0);
        Command command = response.getCommand();
        assertEquals("response type", Command.Type.ERROR,
                command.getType());
        int code = Integer.parseInt(command.getParameter(0));
        for (ServerErrorCode errorCode : expected) {
            if (ServerErrorCode.getCode(errorCode) == code) {
                return;
            }
        }
        fail("error code not in expected set: " + command.getType());
    }

    /**
     * This convenience method calls {@link #assertErrorToUser(int, Set,
     * Broadcast)} with a singleton set, for cases where only a single error
     * code is allowed by the specification.
     *
     * @param userId
     *         the user's ID
     * @param expected
     *         the allowed error code
     * @param actual
     *         the broadcast to check against
     */
    public static void assertErrorToUser(int userId, ServerErrorCode expected,
            Broadcast actual) {
        assertErrorToUser(userId, Collections.singleton(expected), actual);
    }

    /**
     * Prevents the instantiation of any {@code TestUtil} objects, which would
     * be nonsensical.
     */
    private TestUtil() {
    }
}
